// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
//  Giuseppe Vacanti (cosine science & computing)
//  October 23, 2001
// 
//   $Id: Observation.cc,v 1.2 2003/08/26 14:46:43 gvacanti Exp $

#include<Observation.h>

#include<OdfingestUtilities.h>
#include<Dal.h>
#include<Directory.h>
#include<errstr.h>
#include<OdfIcd.h>
#include<SmartDal>
#include<SasDalUtils.h>
#include<TimeRows.h>
#include<ExceptionErrorHandler.h>
#include<OalServer.h>

#include<iostream>
#include<fstream>
#include<functional>
#include<algorithm>
#include<sstream>
#include<string>
#include<cstdlib>
#include<map>
#include<iterator>

namespace {


  typedef std::pair<std::string, std::string> pstring;

  pstring first_and_last_time(const std::string & fname){
    using namespace SmartDal;
    using namespace SasDalUtils;
    DataSetH set = open_set(fname, Read, Low);
    TableH tab = set->table(0);
    ColumnH col = tab->column("TIME");
    ColumnDataH cd = col->data();    
    return std::make_pair(get_string_row(cd,0),
			  get_string_row(cd,tab->rows() - 1));
  }

  // STime with a boolean to indicate if the value has been assigned
  // or not,
  typedef std::pair<STime,bool> STimeH;

  struct TimeBounds {
    STimeH earliest;
    STimeH latest;
  };

  Msg::Layer layer = Msg::App;

  class InstrumentIs {
  public:
    InstrumentIs(const OdfIngest::Instrument & i) : instrument_(i) {}
    bool operator()(const OdfIngest::Exposure & e) const {
      return e.instrument() == instrument_;
    }
  private:
    const OdfIngest::Instrument instrument_;
  };

};
			  


// Constructor makes use of an Odf returned by the OAL.
// The observation is constructed by asking the OAL: exposures, IPPVs
// etc. 
OdfIngest::Observation::Observation(const Odf * odf)
  :
  minimal_exposure_duration_(0),
  directory_(odf->directoryName()),
  identifier_(odf->proposalId() + odf->observationId()),
  revolution_(odf->orbitNr()),
  start_(odf->observationStartTime()),
  end_(odf->observationEndTime()),
  proposal_(odf->proposalInfo()) {
  

  // datasets copied over
  const std::list<OdfSet> & sn = odf->setNames();
  std::copy(sn.begin(), sn.end(), std::back_inserter(sets_));

  // instruments copied over
  typedef std::vector<XmmInstrument *>::const_iterator VI;
  for(VI i = xmmPayLoad.instruments().begin();
      i != xmmPayLoad.instruments().end();
      ++i){

    XmmInstrument * ins = *i;
    
    // ERM is not interesting, and it is never listed in a summary
    // file.
    if(ins->id() == Xmm::ERM)
      continue;
    
    Instrument oins(ins);
    instruments_.push_back(oins);
    
    // exposures copied over 

    // The OAL interface knows how many exposures there are by
    // mode, then each exposure can be extracted by number. This
    // is not a handy interfaces, and here we transform the
    // exposures as knows to the OAL in objects we can easily
    // iterate over.
    const std::vector<Xmm::DataMode> &dm = ins->scienceEngineeringDataModes();
    typedef std::vector<Xmm::DataMode>::const_iterator MI;    
    for(MI k = dm.begin(); k != dm.end(); ++k){
      // How many exposures of a given data mode and instrument?
      unsigned ne = odf->exposures(ins->id(), *k);
      // For each of those we copy the data we need.
      for(unsigned l = 1 ; l <= ne ; ++l){
	exposures_.push_back(
			     Exposure(*(odf->exposure(ins->id(), *k, l)),
				      oins, this));
      }
    }
  }
}

// Constructor scans the ODF directory
// The observation is constructed from scratch, by opening each
// individual file, collecting exposure info, and re-organizing into
// proper exposures.
OdfIngest::Observation::Observation(const std::string & dirname, double minimal_exposure_duration)
  :
  minimal_exposure_duration_(minimal_exposure_duration),
  directory_(dirname),
  start_(STime::latest),
  end_(STime::earliest)
{

  init_instruments();
    
  //----------------------------------------------------------------------
  // Identify how many data sets there are.
  errstr << "Looking for ODF datasets in " << directory_
	 << Sas::message(layer, Msg::Sparse);

  Directory d(directory_);
  d.filter("????_??????????_???????????.???");
  std::transform(d.begin(), d.end(), back_inserter(sets_),
		 mem_fun(&DirEntry::name));
  
  assert(sets_.size() > 0); // FIXME: get rid of the assertion.

  errstr << "There are " << sets_.size() << " datasets."
	 <<  Sas::message(layer, Msg::Sparse);


  //  if(errstr.verbosity(layer) > Msg::Sparse){
    errstr << "Datasets:\n";
    std::transform(sets_.begin(), sets_.end(),
		   ostream_iterator<std::string>(errstr, "\n"),
		   mem_fun_ref(&OdfDataSet::basename));
    Sas::message(layer, Msg::Verbose);
    //  }
  //----------------------------------------------------------------------

    
  //----------------------------------------------------------------------  
  // Determine from the periodic HK the maximum allowed DATE-END for
  // each instrument.
  // Use:
  //	PEH file for MOS and OM
  //	PFH file for RGS
  // 	PMH file for PN

  std::map<Xmm::Instrument, TimeBounds> hk_time_bounds;    
  typedef std::vector<XmmInstrument *>::const_iterator VI;
  for(VI i = xmmPayLoad.instruments().begin();
      i != xmmPayLoad.instruments().end(); ++i){
    XmmInstrument * ins = *i;
    // ERM is not interesting, and it is never listed in a summary
    // file.
    if(ins->id() == Xmm::ERM)
      continue;
    hk_time_bounds[ins->id()];
    hk_time_bounds[ins->id()].earliest.second = false;
    hk_time_bounds[ins->id()].latest.second = false;    
  }

  d.filter("????_??????????_??X00000PEH.FIT|????_??????????_??X00000PFH.FIT|????_??????????_??X00000PMH.FIT");

  typedef Directory::const_iterator LI;
  for(LI j = d.begin(); j != d.end(); ++j){
    OdfIngest::OdfDataSet ds((*j)->name());
    pstring p = first_and_last_time((*j)->name());
    TimeBounds tb;
    try {
      SetSTimeException xx;
      tb.earliest.first = STime(p.first);
      tb.earliest.second = true;
    } catch (STime::InvalidTime e) {
      errstr << "The earliest TIME entry in " << ds.name()
	     << " is not a valid time string: "
	     << p.first << Sas::warning("InvalidTimeString");
      tb.earliest.second = false;
    }
    try {
      SetSTimeException xx;
      tb.latest.first = STime(p.second);
      tb.latest.second = true;
    } catch (STime::InvalidTime e) {
      errstr << "The last TIME entry in " << ds.name()
	     << " is not a valid time string: "
	     << p.first << Sas::warning("InvalidTimeString");
      tb.latest.second = false;
    }
    hk_time_bounds[ds.instrument()] = tb;
    errstr << "The earliest HK time related to "
	   << Xmm::toString(ds.instrument()) << " is "
	   << (tb.earliest.second ? tb.earliest.first.toString(0) : " not defined")
	   << Sas::message(layer, Msg::Sparse);
    errstr << "The latest HK time related to "
	   << Xmm::toString(ds.instrument()) << " is "
	   << (tb.latest.second ? tb.latest.first.toString(0) : " not defined")
	   << Sas::message(layer, Msg::Sparse);
  }

  //----------------------------------------------------------------------
  //Take the first available dataset, and from that determine the
  //revolution number and the observation identifier.

  const OdfDataSet & top = sets_.front();
  revolution_ = top.revolutionNr();
  identifier_ = top.proposalId();

  errstr << "Identified the following observation: "
	 << revolution_ << "_" << identifier_
	 << Sas::message(layer, Msg::Sparse);

  stringstream sumasc;
  sumasc << setw(4) << setfill('0') << revolution_ << "_"
	 << identifier_ << "_SCX00000SUM.ASC";
  
  d.filter(sumasc.str());
  if(d.size() == 1){
    std::string name = (*d.begin())->name();
    const Odf * odf = oalServer.newOdf(name);
    const IPPVCollection & config = odf->configurationParams();
    if(config.size() > 0){
      errstr << "Global configuration parameters:\n";
      for(IPPVCollection::const_iterator K = config.begin();
	  K != config.end(); ++K){
	errstr << (*K).first << " = " << (*K).second << "\n";
	insert_ippv(Ippv((*K).first, (*K).second));
      }
      errstr << Sas::message(layer, Msg::Noisy);
    }

    ifstream is(name.c_str());
    std::string line;
    // Goes through the file until it reaches the PROPOSAL section
    while(getline(is,line) && line.substr(0,8) != "PROPOSAL" && is.good());
    if(!is.eof()){
      the_rest_ << line << "\n";
      while(!is.eof()){
	getline(is,line);
	the_rest_ << line << "\n";
      }
    } else {
      Sas::warning("MissingData", "No PROPOSAL record found. This does not stop you from analyzing your data. If it causes an error at a later stage in the analysis, consider adding your own fake proposal record to the SUM.SAS file.");
    }
  }


  //----------------------------------------------------------------------
  // Go through all the datasets and build a map:
  //    exposure-identifier vs data sets

  typedef std::map<std::string, OdfDataSets,
    std::less<std::string> > EMap;
    
  EMap expos;
  for(OdfDataSets::iterator k = sets_.begin();
      k != sets_.end(); ++k){
    if(k->revolutionNr() != revolution_ || k->proposalId() != identifier_){
      errstr << "The following dataset does not belong to the observation I'm looking at and will be ignored: "
	     << k->basename() << Sas::warning("ExtraneousDataSet");
      continue;
    }
    switch(k->scheduled_status()){
    case OdfDataSet::Scheduled:
    case OdfDataSet::Unscheduled:
      switch(k->instrument()) {
      case Xmm::EMOS1:
      case Xmm::EMOS2:
      case Xmm::EPN:
      case Xmm::RGS1:
      case Xmm::RGS2:
      case Xmm::OM:
	expos[k->exposure_identifier()].push_back(*k);
	break;
	// In case another instrument is ever added :-)
      default:
	errstr << "Do not know about this instrument: " << k->instrument()
	       << Xmm::toString(k->instrument())
	       << Sas::warning("UnknownInstrument");
	break;
      }
      break;
      // Exposures that do not have a scheduled flag ('X') do
      // not contribute to the definition of the structure of
      // the observation in terms of exposures.
    default:
      errstr << "The dataset " << k->basename()
	     << " does not contribute to the structure of the observation."
	     << Sas::message(layer, Msg::Noisy);
    }
  }
  

  //----------------------------------------------------------------------
  // Determine start/stop time for each exposure.
  //
  // We go through all the exposures identified, and try to determine
  // the start/stop time for each of them. Mainly we'll be using the
  // DATE-OBS and DATE-END.
  
  // For OM the dates contained in the science files are known not
  // to be reliable. The correct exposure start/stop times can be
  // found in the NPHK file. One should use an approximate start
  // time (from the science files) as the entry point into the NPHK
  // data.

  // For some types of files, DATE-END can be hours off. In these
  // cases DATE-END will be set to be DATE-OBS +
  // minimal_exposure_duration. See the code for the details.
  for(EMap::iterator k = expos.begin(); k != expos.end(); ++k){
    const std::string & exid = (*k).first;
    OdfDataSets & sv = (*k).second;
    
    errstr << "Identified the exposure " << exid
	   << Sas::message(layer, Msg::Sparse);
    //    if(errstr.verbosity(layer) >= Msg::Verbose){
      errstr << "It contains the following datasets:\n";
      std::transform(sv.begin(), sv.end(),
		     ostream_iterator<std::string>(errstr, "\n"),
		     mem_fun_ref(&OdfDataSet::basename));
      errstr << Sas::message(layer, Msg::Noisy);
      //    }
    
    const OdfDataSet & top = sv.front();
    Instrument instrument(top.instrument());
    bool escheduled = top.scheduled();
    unsigned counter = top.exposureNr();
    
    //----------------------------------------------------------------------
    // Extract the available DATE-OBS, sort them, take the earliest.
    //
    // DATE-OBS can mostly be trusted, but we know that DATE-END
    // cannot. This is why we may have instrument-specific algorithms
    // to refine the start/stop estimates based on DATE-XXX.
    std::vector<STime> vt;
    for(OdfDataSets::iterator k = sv.begin();
	k != sv.end(); ++k){
      OdfDataSet & ds = *k;
      try {
	vt.push_back(ds.start());
	errstr << "Start time for " << ds.basename() << ": "
	       << ds.start().toString(0) << Sas::message(layer, Msg::Verbose);
      } catch (OdfIngest::OdfDataSet::HasNoTime) {
	errstr << "The data set " << ds.basename()
	       << " does not have a valid DATE-OBS, and will not be used to estimate the exposure start time."
	       << Sas::warning("InvalidTimeString");
      }
    }
    
    // If there are no good start times, make use of the one provided
    // by the HK, if available, and warn.
    STime estart;
    if(vt.size() > 0){
      std::sort(vt.begin(), vt.end());
      estart = vt[0]; // candidate start of exposure
    } else {
      errstr << "For " << exid << " no valid DATE-OBS could be identified in the science files."
	     << Sas::message(layer, Msg::Sparse);
      if(hk_time_bounds[instrument.identifier()].earliest.second){
	estart = hk_time_bounds[instrument.identifier()].earliest.first;
	errstr << "Using the start time as calculated from the HK for " << exid
	       << ": " << estart.toString(0) << Sas::message(layer, Msg::Sparse);
      } else
	errstr << "Cannot identify any reasonable start time for " << exid
	       << Sas::error("NoStartTime");
    }
    
      
    //----------------------------------------------------------------------
    // Extract the available DATE-END, sort them, take the latest.
    //    vt.erase(vt.begin(), vt.end());
    vt.clear();
    for(OdfDataSets::iterator k = sv.begin();
	k != sv.end(); ++k){
      OdfDataSet & ds = *k;
      switch(ds.dataMode()) {
      case  Xmm::DISCARDED_LINES_DATA:
      case Xmm::DATA_DIAGNOSTIC:
	errstr << "The data set " << ds.basename() << " is of mode "
	       << ds.dataMode() << ". DATE-END for this type of data sets is not reliable and it shall not be used. Using DATE-END = DATE-OBS + "
	       << minimal_exposure_duration_ << " s instead."
	       << Sas::message(layer, Msg::Verbose);
	vt.push_back(estart + minimal_exposure_duration_);
	break;
      default:
	try {
	  vt.push_back(ds.end());
	  errstr << "End time for " << ds.basename() << ": "
		 << ds.end().toString(0) << Sas::message(layer, Msg::Verbose);	  
	} catch (OdfIngest::OdfDataSet::HasNoTime) {
	  errstr << "The data set " << ds.basename()
		 << " does not have a valid DATE-END, and will not be used to estimate the exposure end time."
		 << Sas::warning("InvalidTimeString");
	}
      }
    }
    
    // If there are no good end times, make use of the one provided
    // by the HK, if available, and warn.
    STime eend;
    if(vt.size() > 0){
      std::sort(vt.begin(), vt.end());
      std::reverse(vt.begin(), vt.end());
      eend = vt[0]; // candidate end of exposure
    } else {
      errstr << "For " << exid << " no valid DATE-END could be identified in the science files."
	     << Sas::message(layer, Msg::Sparse);
      if(hk_time_bounds[instrument.identifier()].latest.second){
	estart = hk_time_bounds[instrument.identifier()].latest.first;
	errstr << "Using the end time as calculated from the HK for " << exid
	       << ": " << eend.toString(0) << Sas::message(layer, Msg::Sparse);
      } else
	errstr << "Cannot identify any reasonable end time for " << exid
	       << Sas::error("NoEndTime");
    }
    

    // For OM the actual exposure start/stop times must be looked for
    // in the NPHK file.
    //----------------------------------------------------------------------
    if(instrument.identifier() == Xmm::OM){

      errstr << "Trying to refine the exposure start/stop times for "
	     << exid << Sas::message(layer, Msg::Sparse);
      errstr << exid << " Candidate start/stop " << estart.toString(0) << "/"
	     << eend.toString(0) << Sas::message(layer, Msg::Verbose);

      std::string omnphk = OdfIcd::hkname(revolution_,
					  identifier_,
					  Xmm::OM,
					  Xmm::NON_PERIODIC_HOUSEKEEPING);
      d.filter(omnphk);
      if(d.size() != 1){
	errstr << "I could not find the file " << omnphk
	       << " while trying to determine the start/stop dates for "
	       << exid << ". Will use the candidate values extracted from the science datasets."
	       << Sas::warning("InexactExposureStartStopTimes");
      } else {
	using namespace SmartDal;
	DataSetH set = SmartDal::open_set(d.front()->name(), Read, Low);
	TableH tab = set->table(0);
	TimeRows tr(tab);
	TimeRows::const_iterator mark = tr.lower_bound(estart);
	if(mark == tr.end())
	  errstr << "Trying to find an entry point in " << omnphk
		 << " around " << estart << " and could not find any. Will use the candidate start/stop values extracted from the science datasets."
		 << Sas::warning("InexactExposureStartStopTimes");
	else {
	  STime refined_estart, refined_eend;
	  ColumnDataH sid = tab->column("SID")->data();
	  int16 * v = sid->int16Data();
	  bool have148 = false;
	  bool have149 = false;
	  for(TimeRows::const_iterator k = mark; k != tr.end(); ++k){
	    if(have148 && have149)
	      break;
	    if(v[tr.index(k)] == 148){
	      try{
		errstr << exid
		       << ": SID 148 (start of exposure) found at record "
		       << tr.index(k) << " "
		       << tr.totime(k).toString(0)
		       << Sas::message(layer, Msg::Verbose);
		refined_estart = tr.totime(k);
		have148 = true;
	      } catch (...) {
		errstr << exid << ": SID 148 found at record " << tr.index(k)
		       << " of " << omnphk << " is not a valid time string."
		       << Sas::warning("InvalidTimeString");
	      }
	    }
	    if(v[tr.index(k)] == 149){
	      try{
		errstr << exid
		       << ": SID 149 (end of exposure) found at record "
		       << tr.index(k) << " "
		       << tr.totime(k).toString(0)
		       << Sas::message(layer, Msg::Verbose);
		refined_eend = tr.totime(k);
		have149 = true;
	      } catch (...) {
		errstr << exid << ": SID 149 found at record " << tr.index(k)
		       << " of " << omnphk << " is not a valid time string."
		       << Sas::warning("InvalidTimeString");
	      }
	    }
	    
	  }
	  if(!have148 && !have149){
	    errstr << exid << ": Could not find both SID148 and SID149 events in "
		   << omnphk << ". This means that the start/stop times cannot be refined." << Sas::warning("InexactExposureStartStopTimes");
	  } else {
	    if(refined_estart > refined_eend)
	      errstr << exid << ": The refined start/stop times found in the housekeeping appear to be wrong (stop < start). " << refined_estart.toString(0) << "/"
		     << refined_eend.toString(0) << ". This means that the start/stop times cannot be refined." << Sas::warning("InexactExposureStartStopTimes");
	    else {
	      estart = refined_estart;
	      eend = refined_eend;
	    }
	  }
	}
      }
    }
    

    errstr << "Start/stop time for " << exid << ": " << estart.toString(0)
	   << " " << eend.toString(0) << Sas::message(layer, Msg::Sparse);
    
    exposures_.push_back(Exposure(counter,
				  escheduled,
				  estart,
				  eend,
				  instrument,
				  sv,
				  this)
			 );	
    
  }

  // Determine the observation start/end times
  if(exposures_.size() > 0){
    std::vector<STime> vt;
    for(Exposures::const_iterator K = exposures_.begin(); K != exposures_.end();
	++K)
      vt.push_back((*K).start_time());
    std::sort(vt.begin(), vt.end());
    start_ = vt[0];
    vt.clear();
    for(Exposures::const_iterator K = exposures_.begin(); K != exposures_.end();
	++K)
      vt.push_back((*K).end_time());
    std::sort(vt.begin(), vt.end());
    std::reverse(vt.begin(), vt.end());
    end_ = vt[0];
  } else
    errstr << "There are no exposures in the ODF." << Sas::warning("NoExposures");
  
}


std::string OdfIngest::Observation::directory() const {
  return directory_;
}

std::string OdfIngest::Observation::identifier() const {
  return identifier_;
}

unsigned OdfIngest::Observation::revolution() const {
  return revolution_;
}

const STime OdfIngest::Observation::start_time() const {
  return start_;
}

const STime OdfIngest::Observation::end_time() const {
  return end_;
}

const OdfIngest::OdfDataSets & OdfIngest::Observation::sets() const {
  return sets_;
}

const OdfIngest::Instruments & OdfIngest::Observation::instruments() const {
  return instruments_;
}

const OdfIngest::Exposures & OdfIngest::Observation::exposures() const {
  return exposures_;
}

unsigned long OdfIngest::Observation::count_exposures(const Instrument & ins) const {
  return std::count_if(exposures_.begin(), exposures_.end(), InstrumentIs(ins));
}
		       
  

OdfIngest::Exposures & OdfIngest::Observation::exposures() {
  return exposures_;
}

const ProposalInfo & OdfIngest::Observation::proposal() const {
  return proposal_;
}

bool OdfIngest::Observation::is_slew() const {
  // FIXME:
  // SLEWS have ids like: 9RRRR#####
  // Unscheduled observations have ids like: 99999#####
  // I (and the ground segment) rely on the fact that we are unlikely
  // to reach 9999 revolutions :) A more accurate test would be
  // observation_id.substring(0,5) == "9RRRR"
  return (identifier_[0] == '9' && identifier_[1] != '9');
}
  

void OdfIngest::Observation::complete_summary(ostream & os) const {
  os << the_rest_.str() << "\n";
}

void OdfIngest::Observation::init_instruments() {
  typedef std::vector<XmmInstrument *>::const_iterator VI;
  for(VI i = xmmPayLoad.instruments().begin();
      i != xmmPayLoad.instruments().end();
      ++i){
    
    XmmInstrument * ins = *i;
    
    // ERM is not interesting, and it is never listed in a summary
    // file.
    if(ins->id() == Xmm::ERM)
      continue;
    
    instruments_.push_back(Instrument(ins));
  }
}
