// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// August 2001
//
// $Id: Observation.h,v 1.2 2003/08/26 14:46:43 gvacanti Exp $
//
#ifndef OBSERVATION_H
#define OBSERVATION_H

#include<Odf.h>
#include<Xmm.h>

//#include<Hkparameter.h>

#include <vector>
#include <string>
#include <map>
#include <iosfwd>

class HkNamer;
class Hkparameter;

namespace OdfIngest {

  // Wrapper classes around existing facilities. The main goal of
  // these classes is to allow one to manipulate more than one ODF
  // at a time. Because the ODF can be had as a data structure in
  // memory, it becomes easier to add/remove items as required.

  class Instrument;
  class Exposure;
  class Observation;
  class OdfDataSet;

  typedef std::vector<OdfDataSet> OdfDataSets;
  typedef std::vector<Exposure> Exposures;
  typedef std::vector<Instrument> Instruments;
  typedef Xmm::DataMode DataMode;
  typedef std::vector<Xmm::DataMode> DataModes;
  typedef IPPV Ippv;
  typedef std::vector<IPPV> Ippvs;

  //----------------------------------------------------------------------
  class OdfDataSet : public OdfSet {
  public:
      
    // exception class
    struct HasNoTime {};
      
    enum ScheduledStatus { Scheduled = 0, Unscheduled, NotApplicable };
      
    OdfDataSet(const std::string & name);
    OdfDataSet(const OdfSet & odfset);
    const STime & start() const throw(HasNoTime);
    const STime & end() const throw(HasNoTime);
    bool has_time() const; // must have both times
    std::string exposure_identifier() const;
    OdfDataSet::ScheduledStatus scheduled_status() const;
      
  private:
    void init_time();
    std::pair<STime, bool> start_, end_;
  };
  
  ostream & operator<<(ostream &, const OdfDataSet &);
  
  //----------------------------------------------------------------------  
  class Instrument {
  public:

    Instrument(const XmmInstrument * instrument);
    Instrument(Xmm::Instrument instrument);
    Instrument(const std::string & nickname);
    std::string name() const;
    Xmm::Instrument identifier() const;
    std::string nickname() const;
    std::string configuration() const;
    const DataModes & science_eng_data_modes() const;
    const DataModes & science_data_modes() const;
    void print(ostream & os = std::cout) const;
    bool operator==(const Instrument & rhs) const;
    bool operator!=(const Instrument & rhs) const;


  private:
    mutable std::map<Xmm::Instrument, std::string> nicknames_;
    mutable std::map<std::string, Xmm::Instrument> rnicknames_;
    mutable std::map<Xmm::Instrument, std::string> configuration_names_;
    const XmmInstrument * xmmInstrument_;

    void init_names();

  };

  //----------------------------------------------------------------------
  class IppvStorage {
  public:
    IppvStorage();

    const Ippvs & ippvs() const;
    Ippvs & ippvs();
    
    bool remove_ippv(const std::string& name);
    bool has_ippv(const std::string& name) const;
    bool insert_ippv(const Ippv & ippv);
    Ippv get_ippv(const std::string& name) const;


  private:
    Ippvs ippvs_;

  };

  //----------------------------------------------------------------------
  class Exposure : public IppvStorage {

  public:
    Exposure(const ::Exposure *, const Instrument &,
	     const Observation * = 0);
    Exposure(const ::Exposure &, const Instrument &,
	     const Observation * = 0);
    Exposure(unsigned counter, bool scheduled, const STime& start,
	     const STime& end, const Instrument &,
	     const OdfDataSets &,
	     const Observation * = 0);
    unsigned counter() const;
    bool scheduled() const;
    const STime & start_time() const;
    const STime & end_time() const;
    const Instrument & instrument() const;
    std::string identifier(bool verbose = false) const;
    bool can_extract_housekeeping_parameters() const;

    const Observation * observation() const;

    void print_to(ostream & os) const;

    //    void extract_hkparameters();

  private:

    void init_data_modes();
    void init_calibration_modes();
    void determine_camera_mode();
    void modes_to_ippvs();

    STime start_, end_;
    Instrument instrument_;
    unsigned counter_;
    bool scheduled_;
    OdfDataSets sets_;
    const Observation * observation_;
    std::map<unsigned, Xmm::DataMode> data_modes_;
    std::map<unsigned, Xmm::Mode> calibration_modes_;
    Xmm::Mode camera_mode_;


  };

  //----------------------------------------------------------------------
  class Observation : public IppvStorage {

  public:

    Observation(const Odf * odf);
    Observation(const std::string & dirname, double minimal_exposure_duration);
    std::string directory() const;
    std::string identifier() const;
    unsigned revolution() const;
    const STime start_time() const;
    const STime end_time() const;
    const OdfDataSets & sets() const;
    const Instruments & instruments() const;
    const Exposures & exposures() const;
    Exposures & exposures();
    unsigned long count_exposures(const Instrument &) const;
    const ProposalInfo & proposal() const;
    bool is_slew() const;
    // Untidy but necessary: some parts of the original summary file
    // are just copied verbatim. There are no high level functions for
    // them. Even the proposal info cannot be reverse-engineered back
    // into a format that can then be read back via the OAL.
    void complete_summary(ostream & os) const;

    
  private:

    Observation(const Observation &);
    Observation & operator=(const Observation &);
    double minimal_exposure_duration_;
    std::string directory_, identifier_;
    unsigned revolution_;
    STime start_, end_;
    OdfDataSets sets_;
    Instruments instruments_;
    Exposures exposures_;
    ProposalInfo proposal_;
    stringstream the_rest_;

    void init_instruments();
  };

  std::ostream & operator<<(std::ostream & os,
			    const Instrument & ins);

  template<class T> void extract_hkparameters(Exposure & exp,
					      std::vector<Hkparameter *> & hkv,
					      double hktimeoffset,
					      const T & hknamer);

  void translate_hkparameters(Exposure & exp);

};

#endif
