// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// June 2000
//

#include<cstdlib>
#include<fstream>
#include<string>
#include<iterator>
#include<algorithm>
#include<iomanip>

// ios_base should be in <ios>
//#include <ios>

#include<param.h>
#include<errstr.h>
#include<metatask.h>
#include<ExceptionErrorHandler.h>

#include<Observation.h>
#include<OalServer.h>

#include<Hkparameter.h>
#include<HkNamer.h>

namespace {
  std::string signature() {
    return std::string(Info::Task::current()->name()) + " ("
      + string(Info::Task::current()->package()) + "-"
      + string(Info::Task::current()->version()) + ") ["
      + string(Info::Task::current()->release()) + "]";
  }

  double max_hk_packet_delta = 64. ; // maximum time granularity of HK packets

};

#include "nodfingest.h"

void nodfingest () {
    
  using namespace OdfIngest;

  std::string odfdir(".");
  ostringstream sf;
  sf.exceptions(ios_base::eofbit);

  double hktimeoffset = realParameter("hktimeoffset");

  const double minimal_exposure_duration = 2. * hktimeoffset + max_hk_packet_delta;
  const unsigned hktolerate = intParameter("hktolerate");

  ostringstream sumfile;

  if(std::getenv("SAS_ODF"))
    odfdir = std::string(std::getenv("SAS_ODF"));

  try {
    SetExceptionErrorHandler x;
    OdfIngest::Observation obs(odfdir,minimal_exposure_duration);

    // Create the canonical summary file name
    sumfile << setw(4) << setfill('0') << obs.revolution() << "_"
	    << obs.identifier() << "_SCX00000SUM.SAS";

    sf << "// ----------------------------------\n";
    sf << "// XMM-Newton Science Analysis System\n";
    sf << "// ----------------------------------\n";
    sf << "// Summary file by " << signature() << "\n";
    sf << "// Created on: " << STime("now").toString(0) << "\n";
    sf << "//\n";
    // PATH may be absent
    if(booleanParameter("writepath"))
      sf << "PATH " << obs.directory() << '\n';
    
    if(obs.is_slew())
      sf << "SLEW\n";
    else
      sf << "OBSERVATION\n";
    sf << obs.identifier() << " / Observation identifier\n";
    sf <<  setw(4) << setfill('0') << obs.revolution() << " / Revolution number\n";
    sf << obs.start_time().toString(0) << " / Start date\n";
    sf << obs.end_time().toString(0) <<  " / Stop date\n";
    
    sf << "// Data sets in the ODF\n";
    sf << "FILES\n";
    sf << obs.sets().size() << "\n";
    std::transform(obs.sets().begin(), obs.sets().end(),
		   ostream_iterator<std::string>(sf, "\n"),
		   mem_fun_ref(&OdfDataSet::basename));
    
    sf << "// Global configuration parameters\n";
    sf << "CONFIGURATION\n";
    Ippvs ip = obs.ippvs();
    sf << ip.size() << "\n";
    for(Ippvs::const_iterator K = ip.begin(); K != ip.end(); ++K)
      sf << (*K).name() << " = " << (*K).value() << "\n";
    

    Instruments ins = obs.instruments();
    Exposures es = obs.exposures();
    
    for(OdfIngest::Instruments::const_iterator II = ins.begin();
	II != ins.end(); ++II){
      sf << "INSTRUMENT\n";
      sf << (*II).nickname() << "\n";
      sf << "Y\n";
      
      sf << obs.count_exposures(*II) << "\n";
      
      typedef OdfIngest::Exposures::iterator EI;
      HkNamer hknamer;
      for(EI i = obs.exposures().begin(); i != obs.exposures().end(); ++i){
	OdfIngest::Exposure e = *i;
	if(e.instrument() != *II)
	  continue;
	if(e.end_time() - e.start_time() < minimal_exposure_duration){
	  errstr << "The exposure " << e.identifier()
		 << " is shorter than " << minimal_exposure_duration
		 << "s. No HK parameters will be extracted."
		 << Sas::warning("ShortExposure");
	  continue;
	}
	if(e.can_extract_housekeeping_parameters()){
	  std::vector<Hkparameter *> hkv =
	    Hkparameter::get_parameters(e.instrument().identifier(), hktolerate);
	  OdfIngest::extract_hkparameters(e, hkv, hktimeoffset, hknamer);
	  OdfIngest::translate_hkparameters(e);
	}
	e.print_to(sf);
      }
    }

    obs.complete_summary(sf);
    
  }
  catch(SasException::Error e){
    Sas::error(e.code, e.message);
  }
  catch(SasException::Fatal e){
    Sas::fatal(e.code, e.message);
  }
  catch(...){
    Sas::fatal("UnknownError", "I detected an error but I cannot tell which one.");
  }

  // If we get here then there were no error

  ofstream summary;
  summary.open(sumfile.str().c_str());
  summary << sf.str();
  summary.close();
  
}
