// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// June 20, 2000
//

#include "IppvExtractorImplementation.h"
#include "IppvImplementation.h"
#include "IppvTranslator.h"
#include <Tracer.h>
#include <errstr.h>
#include <Directory.h>

#include <sstream>

static string int_to_string(const int i, const bool ordinal = false)
{
  PROC
  ostringstream s;
  if(ordinal){
    switch(i){
    case 1:
      s << "1st";
      break;
    case 2:
      s << "2nd";
      break;
    case 3:
      s << "3rd";
      break;
    default:
      s << i << "th";
    }
  } else {      
    s << i;
  }
  RETURN s.str();
}

static map<Xmm::DataMode, vector<Ippv *> > RgsIppvs(const string& prefix)
{
  PROC
  map<Xmm::DataMode, vector<Ippv *> > ippvs;
  vector<Ippv *> v;
  IppvServer is = IppvServer();
  v.push_back(is.ippv("PACKING_SCHEME", prefix + "7303", Dal::DString,
  		      "Packing scheme", Ippv::Periodic, false));
  ippvs[Xmm::FULL_PERIODIC_HOUSEKEEPING] = v;
  v.clear();

  for(unsigned k = 1; k <= 9; k++){
    string s1 = int_to_string(k);

    v.push_back(is.ippv("CCD_" + s1 + "_ACCEPT_THR_C", "ATHRESHC", Dal::Uint16,
			"Acceptance threshold side C for CCD " + s1,
			Ippv::NonPeriodic,
			"CCDID == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_ACCEPT_THR_D", "ATHRESHD", Dal::Uint16,
			"Acceptance threshold side D for CCD " + s1,
			Ippv::NonPeriodic,
			"CCDID == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_UPPER_THR_C", "UTHRESHC", Dal::Uint16,
			"Upper threshold side C for CCD " + s1,
			Ippv::NonPeriodic,
			"CCDID == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_UPPER_THR_D", "UTHRESHD", Dal::Uint16,
			"Acceptance threshold side D for CCD " + s1,
			Ippv::NonPeriodic,
			"CCDID == " + s1, false));

  }
  ippvs[Xmm::DPP_HOUSEKEEPING_1] = v;
  v.clear();

  v.push_back(is.ippv("NUMBER_OF_SLOTS", "CCDSNUM", Dal::Uint16,
		      "Number of readout slots", Ippv::NonPeriodic, false));

  for(unsigned k = 1; k <= 12; k++){
    string s1 = int_to_string(k);
    string s2 = int_to_string(k,true);
    v.push_back(is.ippv("CCD_INDEX_" + s1, "RDOUTSEQ", Dal::Int8,
			"CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_OCB", "OCBMODE", Dal::Int8,
			"On-chip binning factor for the CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_NODES", "READOUT", Dal::Int8,
			"Read-out nodes used fot the CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_REJECT_THR_C", "REJTHRC", Dal::Uint16,
			"Rejection threshold side C for the CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_REJECT_THR_D", "REJTHRD", Dal::Uint16,
			"Rejection threshold side D for the CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
    v.push_back(is.ippv("CCD_" + s1 + "_CSG_PATTERN", "CSGPATT", Dal::Int8,
			"CCD read out in the " + s2 + " slot",
			Ippv::NonPeriodic,
			"#ROW == " + s1, false));
  }
  ippvs[Xmm::DPP_HOUSEKEEPING_2] = v;
  v.clear();
  RETURN ippvs;
}

static map<Xmm::DataMode, vector<Ippv *> > EmosIppvs(const string& prefix)
{
  PROC
  map<Xmm::DataMode, vector<Ippv *> > ippvs;
  vector<Ippv *> v;
  IppvServer is = IppvServer();
  v.push_back(is.ippv("FILTER_WHEEL", prefix + "1257",
		      Dal::DString,
  		      "Name of filter wheel position",
		      Ippv::Periodic, false));
  v.push_back(is.ippv("FILTER_WHEEL_POSITION", prefix + "1317",
		      Dal::Uint16,
		      "Hardware filter wheel position",
		      Ippv::Periodic, false));

  PeriodicIppv<uint16, uint16, &Data::uint16Data> *ip =
    new PeriodicIppv<uint16, uint16, &Data::uint16Data>("FILTER", prefix + "1317", "Filter", false);
  ip->addTranslator(new EmosFilterTranslator<uint16>(prefix));

  v.push_back(ip);


  // Reset time
  v.push_back(is.ippv("CLOCK_RESET_TIME_COARSE", prefix + "1573", Dal::Int32,
		      "Coarse component of the clock reset time",
		      Ippv::Periodic, true));
  v.push_back(is.ippv("CLOCK_RESET_TIME_FINE", prefix + "U573", Dal::Uint16,
		      "Fine component of the clock reset time",
		      Ippv::Periodic, true));
  v.push_back(is.ippv("CLOCK_WRAP_AROUND", prefix + "1044", Dal::Uint16,
		      "Clock wrap around",
		      Ippv::Periodic, false));

  // Mean temparature
  v.push_back(is.ippv("MEAN_TEMPERATURE", prefix + "1311", Dal::Real32,
		      "Mean temperature",
		      Ippv::Periodic, false));

  // readout chain (normal or low gain)
  PeriodicIppv<string, char, &Data::stringData> *ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_1", prefix + "1113", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_2", prefix + "1201", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_3", prefix + "1302", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_4", prefix + "1248", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_5", prefix + "1201", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_6", prefix + "1302", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);
  ip2 =
    new PeriodicIppv<string, char, &Data::stringData>("GAIN_CCD_7", prefix + "1248", "Gain of the read out chain", false);
  ip2->addTranslator(new EmosGainTranslator<string>(prefix));
  v.push_back(ip2);

  // EDU thresholds
  int pstart1 = 1398;
  int pstart2 = 1399;
  int pstep = 2;
  for(unsigned k = 1; k<= 8; k++){
    string s1 = int_to_string(k-1);
    string s2 = int_to_string(k,true);
    string c1 = prefix + int_to_string(pstart1 + (k-1)*pstep);
    string c2 = prefix + int_to_string(pstart2 + (k-1)*pstep);
    v.push_back(is.ippv("EDU_" + s1 + "_LOW_THR_1", c1,
			Dal::Uint16,
			"First low threshold of " + s2 + " EDU",
			Ippv::Periodic, false));
    v.push_back(is.ippv("EDU_" + s1 + "_LOW_THR_2", c2,
			Dal::Uint16,
			"Second low threshold of " + s2 + " EDU",
			Ippv::Periodic, false));
  }
  ippvs[Xmm::PERIODIC_HOUSEKEEPING] = v;
  v.clear();

  // HBR activation status, processing mode, pattern reference,
  // thresholds  
  for(unsigned k = 1; k<= 8; k++){
    string s1 = int_to_string(k);
    string s2 = int_to_string(k-1);
    
    v.push_back(is.ippv("HBR_" + s2 + "_ACTIVATION_STATUS",
			"HBRACTIV",
			Dal::Int8,
			"Activation status for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1, false));
    v.push_back(is.ippv("HBR_" + s2 + "_PROCESSING_MODE",
			"HBRPROC",
			Dal::Uint16,
			"Processing mode for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1, false));
  }
  ippvs[Xmm::HBR_CONFIG_HOUSEKEEPING] = v;
  v.clear();

  for(unsigned k = 1; k<= 8; k++){
    string s1 = int_to_string(k);
    string s2 = int_to_string(k-1);
    
    v.push_back(is.ippv("HBR_PATTERN_REFERENCE_" + s2,
			"PATTERN",
			Dal::Uint16,
			"Pattern reference for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1, false));
    v.push_back(is.ippv("HBR_LOW_THR_" + s2,
			"LTHRESH",
			Dal::Uint16,
			"Low threshold for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1, false));
    v.push_back(is.ippv("HBR_HIG_THR_" + s2,
			"UTHRESH",
			Dal::Uint16,
			"High threshold for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1, false));
  }
  ippvs[Xmm::HBR_THRESHOLD_HOUSEKEEPING] = v;	      

  RETURN ippvs;
}

// The next three static functions code the ODC ICD file name conventions.
// This must be coded here again as the OAL does not give access to it.

static string instrumentToString(const Xmm::Instrument& i)
{
  PROC
  map<Xmm::Instrument, string> m;
  m[Xmm::RGS1] = "R1";
  m[Xmm::RGS2] = "R2";
  m[Xmm::EPN] = "PN";
  m[Xmm::EMOS1] = "M1";
  m[Xmm::EMOS2] = "M2";
  m[Xmm::OM] = "OM";
  m[Xmm::ERM] = "RM";
  if(m.find(i) == m.end())
    errstr << "The instrument " << i <<
      " is not recognized here." << Sas::error("UnknownInstrument");
  RETURN m[i];
}

static string modeToString(const Xmm::DataMode& d)
{
  PROC
  map<Xmm::DataMode, string> m;
  m[Xmm::PERIODIC_HOUSEKEEPING] = "PE";
  m[Xmm::HBR_CONFIG_HOUSEKEEPING] = "HC";
  m[Xmm::HBR_BUFFER_HOUSEKEEPING] = "HB";
  m[Xmm::HBR_THRESHOLD_HOUSEKEEPING] = "HT";
  m[Xmm::EXTRAHEATING_CONF_HOUSEKEEPING] = "EC";
  m[Xmm::THERMAL_HOUSEKEEPING] = "TM";
  m[Xmm::BRIGHT_PIXEL_HOUSEKEEPING] = "PT";
  m[Xmm::MAIN_PERIODIC_HOUSEKEEPING] = "PM";
  m[Xmm::ADD_PERIODIC_HOUSEKEEPING] = "PA";
  m[Xmm::FULL_PERIODIC_HOUSEKEEPING] = "PF";
  m[Xmm::TEMP_PERIODIC_HOUSEKEEPING] = "PC";
  m[Xmm::DPP_HOUSEKEEPING] = "DP";
  m[Xmm::NON_PERIODIC_HOUSEKEEPING] = "NP";
  m[Xmm::DPP_HOUSEKEEPING_1] = "D1";
  m[Xmm::DPP_HOUSEKEEPING_2] = "D2";
  m[Xmm::HK_P1] = "P1";
  m[Xmm::HK_P2] = "P2";
  m[Xmm::HK_P3] = "P3";
  m[Xmm::HK_P4] = "P4";
  m[Xmm::HK_P5] = "P5";
  m[Xmm::HK_P6] = "P6";
  m[Xmm::HK_P7] = "P7";
  m[Xmm::HK_P8] = "P8";
  m[Xmm::HK_P9] = "P9";
  if(m.find(d) == m.end())
    errstr << "The data mode " << Xmm::toString(d) <<
      " is not recognized here." << Sas::error("UnknownDataMode");
  RETURN m[d];
}

static string hkname(const string& directory, const Xmm::Instrument& i,
		     const Xmm::DataMode& d)
{
  PROC
  Directory dir(directory);
  string s("????_??????????_" + instrumentToString(i) +
	   "X00000" + modeToString(d) + "H.FIT");
  dir.filter(s);
  if(dir.size() !=1 ){
//      errstr << "In the directory " << directory <<
//        " I expected to find one and only one dataset matching the pattern " <<
//        s << ". Instead I found " << dir.size() <<
//        Sas::error("IncorrectNumberOfDatasets");
    stringstream s;
    s << "In the directory " << directory <<
      " I expected to find one and only one dataset matching the pattern " <<
      dir.filter() << ". Instead I found " << dir.size();
    throw(NoHousekeepingFile(s.str()));
  }
  RETURN (*dir.begin())->name();
}


// IppvExtractorServer

IppvExtractorServer::IppvExtractorServer()
{
  PROC ENDPROC
}

IppvExtractorServer::~IppvExtractorServer()
{
  PROC ENDPROC
}

IppvExtractor * IppvExtractorServer::extractor(Xmm::Instrument ins,
					       const string& directory ,
					       const string& label )
{
  PROC
    switch(ins) {
    case Xmm::RGS1:
      ALTRETURNV(new Rgs1IppvExtractor(directory, label));
      break;
    case Xmm::RGS2:
      ALTRETURNV(new Rgs2IppvExtractor(directory, label));
      break;
    case Xmm::EMOS1:
      ALTRETURNV(new Emos1IppvExtractor(directory, label));
      break;
    case Xmm::EMOS2:
      ALTRETURNV(new Emos2IppvExtractor(directory, label));
      break;
    case Xmm::EPN:
      ALTRETURNV(new EpnIppvExtractor(directory, label));
      break;
    case Xmm::OM:
      ALTRETURNV(new OmIppvExtractor(directory, label));
      break;
    case Xmm::ERM:
      ALTRETURNV(new ErmIppvExtractor(directory, label));
      break;
    default:
      errstr << "No IPPV extractor available for instrument " <<
	Xmm::toString(ins) << Sas::error("NoIppvExtractorAvailable");      
    }
  RETURN 0;
}

// IppvExtractorCommon

IppvExtractorCommon::IppvExtractorCommon(const Xmm::Instrument instrument,
					 const string& directory ,
					 const string& label )
  : _dir(directory),
    _instrument(instrument),
    _label(label)   
{
  PROC
  ENDPROC
}

unsigned IppvExtractorCommon::ippvs() const
{
  PROC
  unsigned n = 0;
  for(map<Xmm::DataMode, vector<Ippv *> >::const_iterator i = _ippvs.begin();
      i != _ippvs.end(); i++){
    n += (*i).second.size();
  }
  RETURN n;
}

void IppvExtractorCommon::write(ostream& os,
				const STime& start,
				const STime& stop) const
{
  PROC
  for(map<Xmm::DataMode, vector<Ippv *> >::const_iterator i = _ippvs.begin();
      i != _ippvs.end(); i++){
    const vector<Ippv *> v = (*i).second;
    try{ 
      string hk = hkname(_dir, _instrument, (*i).first);
      DataSet *set = dataSetServer->open(hk, Dal::Read, DataSetServer::Low);
      errstr << _label << " Fetching IPPVs from " << hk
	     << " Time range: " << start << " - " << stop
	     << Sas::message(Msg::AppLib, Msg::Sparse);
      { // This scope is needed to make TimeOrderedTable call its destructor before the dataset is cosed.
	TimeOrderedTable tot = TimeOrderedTable(set->table(0));
	for(unsigned k = 0; k < v.size(); k++){
	  v[k]->print(errstr,true);
	  errstr << Sas::message(Msg::AppLib, Msg::Noisy);
	  try {
	    v[k]->fetch(&tot, start, stop);
	  }
	  catch (IppvException e) {
	    errstr << _label << " ";
	    e.print(errstr);
	    errstr << Sas::warning("HousekeepingParameterMissing");
	    v[k]->notAvailable(os);
	    continue;
	  }
	  v[k]->print(os);
	}
      }
      dataSetServer->close(set);
    }
    catch(IppvException e) {
      e.print(errstr);
      errstr << Sas::warning("HousekeepingParameterMissing");
      for(unsigned k = 0; k< v.size(); k++){
	errstr << " " << _label << " : " << v[k]->name()
	       << Sas::warning("HousekeepingParameterMissing");
	v[k]->notAvailable(os);
      }
    }
  }
  ENDPROC
}

// SCR117
map<string,string> IppvExtractorCommon::values(const STime& start,
									const STime& stop) const
{
	PROC
	
	map<string,string> ippvValues;
	
	for(map<Xmm::DataMode, vector<Ippv *> >::const_iterator i = _ippvs.begin();
		i != _ippvs.end(); i++)
	{
		const vector<Ippv *> v = (*i).second;
		try
		{ 
			string hk = hkname(_dir, _instrument, (*i).first);
			DataSet *set = dataSetServer->open(hk, Dal::Read, DataSetServer::Low);
			errstr << _label << " Fetching IPPVs from " << hk
				<< " Time range: " << start << " - " << stop
				<< Sas::message(Msg::AppLib, Msg::Sparse);
			{ // This scope is needed to make TimeOrderedTable call its destructor before the dataset is cosed.
				TimeOrderedTable tot = TimeOrderedTable(set->table(0));
				for(unsigned k = 0; k < v.size(); k++)
				{
					try 
					{
						v[k]->fetch(&tot, start, stop);
					}
					catch (IppvException e) 
					{
						errstr << _label << " ";
						e.print(errstr);
						errstr << Sas::warning("HousekeepingParameterMissing");
						continue;
					}
					
					string val;
					v[k]->value(val);
					
					ippvValues[ v[k]->name() ] = val;
					
				}
			}
			dataSetServer->close(set);
		}
		catch(IppvException e) 
		{
			e.print(errstr);
			errstr << Sas::warning("HousekeepingParameterMissing");
			for(unsigned k = 0; k< v.size(); k++)
			{
				errstr << " " << _label << " : " << v[k]->name()
					<< Sas::warning("HousekeepingParameterMissing");
			}
		}
	}

	RETURN ippvValues;
}


void IppvExtractorCommon::document(ostream& os) const
{
  PROC
  string s("Parameters for " + Xmm::toString(instrument()));
  os << string(s.size(), '=') << endl;
  os << s << endl;
  for(map<Xmm::DataMode, vector<Ippv *> >::const_iterator i = _ippvs.begin();
      i != _ippvs.end(); i++){
    const vector<Ippv *> v = (*i).second;
    s = "HK file type: " + Xmm::toString((*i).first);
    os << string(s.size(), '-') << endl;
    os << s << endl;
    os << "[" << v.size() << " parameters]" << endl;
    os << string(s.size(), '-') << endl;
    for(unsigned k = 0; k < v.size(); k++)
      v[k]->print(os,true);
  }
  ENDPROC
}
      

Xmm::Instrument IppvExtractorCommon::instrument() const
{
  PROC
  RETURN _instrument;
}

// Rgs1IppvExtractor

Rgs1IppvExtractor::Rgs1IppvExtractor(const string& directory ,
				     const string& label )
  : IppvExtractorCommon(Xmm::RGS1, directory, label)
{
  PROC
  _ippvs = RgsIppvs("G"); // The prefix for the RGS1 HK columns is "G"  
  ENDPROC
}

// Rgs2IppvExtractor

Rgs2IppvExtractor::Rgs2IppvExtractor(const string& directory ,
				     const string& label )
  : IppvExtractorCommon(Xmm::RGS2, directory, label)
{
  PROC
  _ippvs = RgsIppvs("L"); // The prefix for the RGS2 HK columns is "L"  
  ENDPROC
}

// Emos1IppvExtractor

Emos1IppvExtractor::Emos1IppvExtractor(const string& directory ,
				       const string& label )
  : IppvExtractorCommon(Xmm::EMOS1, directory, label)
{
  PROC
  _ippvs = EmosIppvs("E"); // The prefix for the EMOS1 HK columns is "E"
  ENDPROC
}

// Emos2IppvExtractor

Emos2IppvExtractor::Emos2IppvExtractor(const string& directory ,
				       const string& label )
  : IppvExtractorCommon(Xmm::EMOS2, directory, label)
{
  PROC
  _ippvs = EmosIppvs("K"); // The prefix for the EMOS2 HK columns is "K"
  ENDPROC
}

// EpnIppvExtractor

EpnIppvExtractor::EpnIppvExtractor(const string& directory ,
				   const string& label )
  : IppvExtractorCommon(Xmm::EPN, directory, label)
{
  PROC
  vector<Ippv *> v;
  IppvServer is = IppvServer();

  // SCR 191
  PeriodicIppv<real32, real32, &Data::real32Data> *ipp =
    new PeriodicIppv<real32, real32, &Data::real32Data>("FILTER", "F1122", "Filter position translated from potentiometer readout", false);
  ipp->addTranslator(new EpnFilterTranslator<real32>());

  v.push_back(ipp);

  v.push_back(is.ippv("FILTER_WHEEL_POSITION",
		      "F1122",
		      Dal::Real32,
  		      "Hardware filter wheel position",
		      Ippv::Periodic,
		      "F1119 == \"STOPPED\"",
		      false
		      ));

  v.push_back(is.ippv("FRAME_TIME_PARAMETER", "F1294", Dal::Int8,
		      "", Ippv::Periodic, false));
  // Reset time
  v.push_back(is.ippv("CLOCK_RESET_TIME_COARSE", "F1030", Dal::Int32,
		      "Coarse component of the clock reset time",
		      Ippv::Periodic, true));
  v.push_back(is.ippv("CLOCK_RESET_TIME_FINE", "FU036", Dal::Uint16,
		      "Fine component of the clock reset time",
		      Ippv::Periodic, true));
  v.push_back(is.ippv("CLOCK_WRAP_AROUND", "F1052", Dal::Uint16,
		      "Clock wrap around",
		      Ippv::Periodic, false));

  // CAMEX gains
  v.push_back(is.ippv("CAMEX_GAIN_CCD1", "F1227", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD2", "F1226", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD3", "F1225", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD4", "F1234", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD5", "F1233", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD6", "F1232", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD7", "F1241", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD8", "F1240", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD9", "F1239", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD10", "F1248", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD11", "F1247", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CAMEX_GAIN_CCD12", "F1246", Dal::DString,
		      "", Ippv::Periodic, false));
			  
  _ippvs[Xmm::MAIN_PERIODIC_HOUSEKEEPING] = v;

  // Thresholds
  v.clear();
  v.push_back(is.ippv("LOWER_THRESHOLD_Q0_CCD0", "F1515", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q0_CCD1", "F1516", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q0_CCD2", "F1517", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q1_CCD0", "F1615", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q1_CCD1", "F1616", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q1_CCD2", "F1617", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q2_CCD0", "F1715", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q2_CCD1", "F1716", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q2_CCD2", "F1717", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q3_CCD0", "F1815", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q3_CCD1", "F1816", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("LOWER_THRESHOLD_Q3_CCD2", "F1817", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CMCORR_Q0", "F1525", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CMCORR_Q1", "F1625", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CMCORR_Q2", "F1725", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("CMCORR_Q3", "F1825", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("QUADRANT_UNDERSAMPLING_Q0", "F1534", Dal::Int8,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("QUADRANT_UNDERSAMPLING_Q1", "F1634", Dal::Int8,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("QUADRANT_UNDERSAMPLING_Q2", "F1734", Dal::Int8,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("QUADRANT_UNDERSAMPLING_Q3", "F1834", Dal::Int8,
		      "", Ippv::Periodic, false));

  // Bright Pixel mode (masked mode)			  
  v.push_back(is.ippv("A1_CMLINE_PIXSET", "F1628", Dal::DString,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("A1_CMLINE_CCD", "F1629", Dal::Int8,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("A1_CMLINE_NUMB", "F1630", Dal::Int8,
		      "", Ippv::Periodic, false));

  _ippvs[Xmm::ADD_PERIODIC_HOUSEKEEPING] = v;

  // HBR
  v.clear();
  for(unsigned k = 1; k <=4; k++){
    string s1 = int_to_string(k);
    string s2 = int_to_string(k-1);
    
    v.push_back(is.ippv("HBR_" + s2 + "_ACTIVATION_STATUS",
			"HBRPROC",
			Dal::Uint16,
			"Activation status for HBR " + s2,
			Ippv::NonPeriodic,
			"HBRID == " + s1,
			true));
  }
  _ippvs[Xmm::HBR_CONFIG_HOUSEKEEPING] = v;
  
  ENDPROC
}

// OmIppvExtractor

OmIppvExtractor::OmIppvExtractor(const string& directory ,
				 const string& label )
  : IppvExtractorCommon(Xmm::OM, directory, label)
{
  PROC
  vector<Ippv *> v;
  IppvServer is = IppvServer();

  v.push_back(is.ippv("EXPOSURE_DURATION", "H5440", Dal::Uint16,
		      "", Ippv::Periodic, false));
  v.push_back(is.ippv("FILTER_WHEEL_POSITION", "H5265", Dal::Int16,
		      "Hardware filter wheel position", Ippv::Periodic, false));	

  PeriodicIppv<int16, int16, &Data::int16Data> *ip =
    new PeriodicIppv<int16, int16, &Data::int16Data>("FILTER", "H5265", "Filter", false);
  ip->addTranslator(new OmFilterTranslator<int16>());
  v.push_back(ip);

  // SCR117
  v.push_back(is.ippv("FLOOD_LED_BIAS", "H5195", Dal::Real32,
		      "LED Status", Ippv::Periodic, false));


  _ippvs[Xmm::PERIODIC_HOUSEKEEPING] = v;
  v.clear();

  ENDPROC
}

// ErmIppvExtractor: needed to define the common Ippv
//
ErmIppvExtractor::ErmIppvExtractor(const string& directory ,
                                      const string& label )
          : IppvExtractorCommon(Xmm::ERM, directory, label)
{
  PROC
  ENDPROC
}

