// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
//
// XMM Science Analysis System
//
// Giuseppe Vacanti (European Space Agency) 
// $Date: 2012/12/18 14:26:13 $
//
// 
#ifndef OdfObject_H
#define OdfObject_H

#include <string>
#include <list>

#include <vector>
#include <map>
#include <set>

#include <STime.h>
#include <Xmm.h>
#include "OdfConstituent.h"
#include "ExposureObject.h"
#include "Dal.h"
#include <Odf.h>

// OdfObject represents an ODF from the point of view of a task
// writing the ODF summmary file.

class OdfObject {
  
 public:

  OdfObject(const string& dir = ".", const bool findinstrumentmodes = true,
	    const bool usehousekeeping = true);
  string identifier() const;
  void writeSummary(ostream& os = cout, const bool writepath = false) const;  
  string canonicalSummaryName() const;
  void insertFile(const string name);
  void removeFile(const string name);
  string icdosf() const;
  const bool& valid() const;

  void checkExpTimeSpan(int pointing,string pointingFile) ;
  void replaceName(int pointing,string pointingFile) ;
  void removeExpFromASC(Xmm::Instrument inst, unsigned int exp, bool schedFlag, bool schedFlagApplicable);
  void writeNumberOfFiles(string ascFileName);
  void ascrename(int pointing,string pointingFile);
  bool modifyDLI(Block *DLIblk, OdfSet odfset, string fileName, DataSet *set, STime startTime, STime stopTime  );
  bool removeExp(OdfSet odfset, string fileName, DataSet *set,Block *blk);
  void modifyODI(Block *ODIblk, STime startTime);
  void checkOMPHEExposures();
  ExposureObject  modifyHDI();

 private:
  
  // Find all ODF constituents in a given directory.
  void scan();

  // Methods to write the various sections of the summary file.
  void observation_rec(ostream& os) const;
  void file_details_rec(ostream& os) const;
  void configuration_details_rec(ostream & os) const;
  void instrument_details_rec(const string& inst,
			      ostream& os) const;
  void exposure_sub_record(const string& inst,
			   const string& exp,
			   ostream& os) const;
  void configuration_sub_record(const string& inst,
				const string& exp,
				ostream& os) const;
  void proposal_summary_rec(ostream& os) const;

  // Given an instrument name and the type of HK file, return the full
  // name of the HK file.
  string hkfile(const string& i, const string& type);


  const list<STime> exposure_times(const string& inst,
				   const string& exp) const;

  vector<STime> exposure_start_end_times(const string& inst,
					 const string& exp) const;
  
  bool is_instrument_active(const string& inst) const;
  unsigned int number_of_exposures_of_instrument(const string& inst) const;

  multimap<string,OdfConstituent,less<string> > file_map_by_inst;
  multimap<string,string,less<string> > exposure_map_by_inst; 
  multimap<string,ExposureObject,less<string> > inst_exp_time_map;

  vector<string> instruments;
  vector<string> file_list;
  vector<Xmm::DataMode> science_data_modes;

  mutable map<string,string,less<string> > long_names;
  mutable map<string,string,less<string> > fits_names;
  
  set<string, less<string> > file_set;
  
  string revolution;
  string proposal;
  string observation_id;
  mutable STime start_time;
  mutable STime end_time;

  string _dir;
  const bool _find_instrument_modes;
  const bool _use_housekeeping;
  bool _valid;

  const double _tpad;
  vector<Xmm::DataMode> _no_ippv_data_modes;

};

#endif
